import 'dart:async';
import 'dart:convert';
import 'dart:math';
import 'dart:typed_data';
// ignore: avoid_web_libraries_in_flutter
import 'dart:html' as html;
import 'package:convert/convert.dart';
import 'package:http/http.dart' as http;
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:http/http.dart';
import 'package:isolated_worker/js_isolated_worker.dart';
import 'package:pointycastle/api.dart';
import 'package:pointycastle/digests/sha256.dart';
import 'package:pointycastle/export.dart';
import 'package:robomotion_app/errors.dart';
import 'package:robomotion_app/srp/client.dart';
import 'package:robomotion_app/srp/standard_groups.dart';
import 'package:robomotion_app/utils/extensions.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import 'package:uuid/uuid.dart';
import 'package:encrypt/encrypt.dart' as enc;
import 'package:fast_rsa/fast_rsa.dart' as rsa;
import 'package:cryptography/cryptography.dart';
import '../encryption.dart';
import 'models/published_flow.dart';
import 'models/queue_message.dart';

StreamController<void> _keyExStreamCtrl = StreamController<void>.broadcast();
StreamController<void> _robotIDStreamCtrl = StreamController<void>.broadcast();

class RError {
  Map? data;
  RError() {
    this.data = null;
  }
}

class Vault {
  final Map vault;
  List<int>? secretKey;
  List<int>? _vaultKey;
  String encSecretKey;
  String encVaultKey;

  Vault({
    required this.vault,
    this.secretKey,
    List<int>? vaultKey,
    this.encSecretKey = '',
    this.encVaultKey = '',
  }) : _vaultKey = vaultKey;

  Future<List<int>> getVaultKey(Map user, String password) async {
    if (_vaultKey != null) {
      return _vaultKey!;
    }

    encVaultKey = vault['enc_vault_key'];
    String pkPem = await Robomotion().getPrivateKey(
      user,
      password,
    );

    final vk = await rsa.RSA.decryptOAEPBytes(
      base64Decode(encVaultKey),
      "",
      rsa.Hash.SHA256,
      pkPem,
    );

    _vaultKey = vk.toList();
    return _vaultKey!;
  }
}

class Robomotion {
  Config? config;
  void Function(QueueMessage data)? onMsgRecv;
  void Function(String message)? onRequestFailed;
  String? _apiURL;
  String? _appURL;
  String? _sid;
  late WebSocketChannel _channel;
  final Map<String, StreamController<dynamic>> _streams = {};

  Robomotion({this.config, this.onMsgRecv}) {
    config ??= Config();
    _apiURL = _getAPIUrl();
    _appURL = _getAppUrl();
    _sid = _getCookie("sid") ?? "5543d0c5-773b-4bca-9a0f-8bc9c195ca5b";
    generateAESKey();

    this.onMsgRecv ??= (msg) {
      if (_streams.containsKey(msg.id)) {
        var payload = jsonDecode(utf8.decode(msg.payload!));
        _streams[msg.id]?.add(payload);
      }
    };
  }

  Future<void> run() async {
    await _connectAppServer();
  }

  Future<Map?> login(String domain, String username, String password) async {
    String nonce = Uuid().v4().replaceAll("-", "") +
        (DateTime.now().millisecondsSinceEpoch / 1000).toString();
    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/auth.challenge',
      body: {
        "username": username,
        "domain": domain,
        "nonce": nonce,
      },
    );

    if (resp == null) {
      return null;
    }

    var client = SRP6Client(
      group: SRP6StandardGroups.rfc5054_4096,
      digest: SHA256Digest(),
      random: _getSecureRandom(),
    );

    var A = await client.generateClientCredentialsAsync(
      Uint8List.fromList(hex.decode(resp["salt"])),
      Uint8List.fromList(utf8.encode(username)),
      Uint8List.fromList(utf8.encode(password)),
    );
    A = A ?? BigInt.zero;

    client.calculateSecret(BigInt.parse(resp["b"], radix: 16));
    var m1 = client.calculateClientEvidenceMessage() ?? BigInt.zero;

    var loginResp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/auth.login',
      body: {
        "username": username,
        "domain": domain,
        "nonce": nonce,
        "M1": m1.toRadixString(16),
        "A": A.toRadixString(16),
      },
    );

    return loginResp;
  }

  SecureRandom _getSecureRandom() {
    var secureRandom = FortunaRandom();
    var random = Random.secure();
    List<int> seeds = [];
    for (int i = 0; i < 32; i++) {
      seeds.add(random.nextInt(255));
    }
    secureRandom.seed(KeyParameter(Uint8List.fromList(seeds)));
    return secureRandom;
  }

  Future<List<PublishedFlow>> getPublishedFlows() async {
    String url = "${this._apiURL}/v1/flows.published.list";
    Map resp = await (_request(
      method: "GET",
      url: url,
      queryParams: {"page": "1", "size": "100", "search": ""},
    ));

    List<PublishedFlow> flows = [];
    for (dynamic f in resp["flows"]) {
      flows.add(PublishedFlow.fromJson(f));
    }
    return flows;
  }

  Future<String?> runFlow(String robotID, String flowID) async {
    String url = "${this._apiURL}/v1/apps.flow.run";
    String? _flowID = flowID;
    RError err = RError();

    final resp = await (_request(
        method: "POST",
        url: url,
        error: err,
        body: {"robot_id": robotID, "app_id": config!._id, "flow_id": flowID}));

    if (err.data != null) {
      config!.robotID = err.data!["robot_id"];
      try {
        var error = jsonDecode(err.data!["error"]);
        if (error.containsKey("payload") &&
            error["payload"].containsKey("flow_id")) {
          _flowID = error["payload"]["flow_id"];
        } else {
          throw Exception(error);
        }
      } catch (ex) {
        throw Exception(err.data!["error"]);
      }
    }

    config!.robotID = resp.containsKey("robot_id") ? resp["robot_id"] : "";
    return _flowID;
  }

  Future<Map?> apiRequest({
    required String method,
    required String endpoint,
    Map? body,
    Map<String, String>? queryParams,
    bool withCredentials = true,
    RError? error,
  }) async {
    if (!endpoint.startsWith("/")) {
      endpoint = "/$endpoint";
    }

    return _request(
      method: method,
      url: "$_apiURL$endpoint",
      body: body!,
      queryParams: queryParams,
      withCredentials: withCredentials,
      error: error,
    );
  }

  Future<Map<dynamic, dynamic>> _request({
    required String? method,
    required url,
    Map<dynamic, dynamic>? body,
    Map<String, String>? queryParams,
    bool withCredentials = true,
    RError? error,
  }) async {
    Map<String, String> headers = {
      'Content-Type': 'application/json; charset=UTF-8',
    };

    if (kDebugMode) {
      headers['Authorization'] = 'Bearer $_sid';
    }
    Uri uri = Uri.parse(url);
    if (queryParams != null) {
      uri = Uri.parse(url).replace(queryParameters: queryParams);
    }
    var request = http.Request(method!, uri);
    request.headers.addAll(headers);

    if (body != null) {
      request.body = json.encode(body);
    }

    Response response = await Response.fromStream(await request.send());

    if (response.body != "") {
      Map<String, dynamic> body = json.decode(response.body);

      if (!body.containsKey("ok") || !body["ok"]) {
        if (body["error"] == ErrAuthFailed) {
          html.window.location.href = Uri.base.origin;
          return Map();
        }

        if (error != null) {
          error.data = body;
          return Map();
        }

        if (onRequestFailed != null) {
          onRequestFailed?.call(body.toString());
          return Map();
        }

        throw new Exception(body.toString());
      }
    }

    Map<String, dynamic> _res = json.decode(response.body);
    return _res;
  }

  Future<void> _connectAppServer() async {
    String url = "$_appURL?id=${config!._id}";
    if (kDebugMode) {
      url += "&sid=$_sid";
    }

    _channel = WebSocketChannel.connect(Uri.parse(url));
    _channel.stream.listen((data) {
      final msg = QueueMessage.fromJson(jsonDecode(data));
      if (msg.encrypted!) {
        msg.payload = decryptPayload(msg.payload!);
      }
      _onMsgRecv(msg).then((_) {
        if (onMsgRecv != null) onMsgRecv!(msg);
      });
    });
  }

  Future<void> _onMsgRecv(QueueMessage msg) async {
    if (msg.msgType == QMT_Error) {
      return;
    }

    if (config!.robotID == "" && msg.msgType == QMT_RobotID) {
      await _robotIDStreamCtrl.stream.first;
    }

    switch (msg.msgType) {
      case QMT_AppRequest:
        final resp = QueueMessage(
            id: msg.id,
            appID: config!._id,
            robotID: config!.robotID,
            msgType: QMT_AppRequest,
            payload: utf8.encode('{"msg": "received"}') as Uint8List?,
            encrypted: true);

        sendAppMsg(resp);
        break;

      case QMT_KeyExchange:
        final pem = msg.payload!;
        robotPubKeyPem = utf8.decode(pem);
        robotPubKey = parsePublicKey(pem);
        _keyExStreamCtrl.add(null);
        break;

      case QMT_RobotID:
        config!.robotID = utf8.decode(msg.payload!);
        _robotIDStreamCtrl.add(null);
        break;
    }
  }

  void sendAppMsg(QueueMessage msg) {
    if (msg.encrypted!) {
      msg.payload = encryptPayload(msg.payload!);
    }
    _channel.sink.add(msg.toString());
  }

  Future<void> keyExchange() async {
    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: config!._id,
      robotID: config!.robotID,
      userID: config!._userID,
      msgType: QMT_KeyExchange,
      payload: null,
      encrypted: false,
    );

    sendAppMsg(msg);
    await _keyExStreamCtrl.stream.first;

    final encAes = rsaEncrypt(robotPubKey, aesKey.bytes);
    msg = QueueMessage(
        id: Uuid().v4(),
        appID: config!._id,
        robotID: config!.robotID,
        msgType: QMT_KeyExchange,
        payload: encAes,
        encrypted: false);

    sendAppMsg(msg);
  }

  String appIn(
    Map payload, {
    appID,
    robotID,
    void Function(dynamic)? onResponse,
  }) {
    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: appID == null ? config!._id : appID,
      robotID: robotID == null ? config!.robotID : robotID,
      userID: config!._userID,
      msgType: QMT_AppIn,
      payload: utf8.encode(jsonEncode({"payload": payload})) as Uint8List?,
      encrypted: true,
    );

    sendAppMsg(msg);

    if (onResponse != null) {
      subscribe(msg.id!, onResponse);
    }

    return msg.id!;
  }

  void subscribe(String id, void Function(dynamic) f) {
    StreamController<dynamic> respStream =
        StreamController<dynamic>.broadcast();
    _streams[id] = respStream;
    respStream.stream.first.then((value) {
      f(value);
      respStream.close();
      _streams.remove(id);
    });
  }

//Isolate
  List<String> _jsScripts = <String>['encryptPayload.js'];
  String _jsGetFunctionName = 'encrypt';

  Future<Uint8List> encrypt(Uint8List data) async {
    await JsIsolatedWorker().importScripts(_jsScripts);
    return await JsIsolatedWorker()
        .run(functionName: _jsGetFunctionName, arguments: data);
  }

  Future<String> _getUploadLink(String file) async {
    String url = "${this._apiURL}/v1/apps.upload.get";
    var resp = await (_request(
        method: "GET",
        url: url,
        queryParams: {"app_id": config!._id, "file": file}));
    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }
    return Uri.decodeFull(resp["url"]);
  }

  Future<void> uploadFile(String file, Uint8List data) async {
    String url = await _getUploadLink(file);
    data = await encrypt(data);
    Uri uri = Uri.parse(url);
    var request = new http.MultipartRequest("PUT", uri);
    request.headers.addAll({"Content-Type": "application/octet-stream"});
    request.files.add(new http.MultipartFile.fromBytes('file', data));
    await request.send();
  }

  Future<String?> getFileURL(String file) async {
    String url = "${this._apiURL}/v1/apps.file.get";
    var resp = await (_request(
        method: "GET",
        url: url,
        queryParams: {"app_id": config!._id, "file": file}));
    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }
    return resp["url"];
  }

  String _getAPIUrl() {
    String ws = this.config!.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0];
    return domain == "robomotion.io" ? "$protocol//api.$domain" : "$ws/api";
  }

  String _getAppUrl() {
    String ws = this.config!.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0] == "https:" ? "wss:" : "ws:";
    return domain == "robomotion.io"
        ? "$protocol//mq.$domain/app"
        : "$protocol//${parts[0].split("//")[1]}.$domain/mq/app";
  }

  Future<Map?> getUser() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/users.get',
    );

    if (resp == null || !resp['ok']) {
      return null;
    }

    return resp['user'];
  }

  Future<Map<String, Vault>?> getVaults() async {
    var resp = await apiRequest(method: 'GET', endpoint: '/v1/vaults.list');

    if (resp == null || !resp['ok']) {
      return null;
    }

    Map<String, Vault> vaults = {};
    for (var v in resp['vaults']) {
      vaults[v['id']] = Vault(vault: v);
    }

    return vaults;
  }

  Future<Map<String, dynamic>?> getVaultItems(String vaultID) async {
    RError error = RError();
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/vaults.items.list',
      queryParams: {'vault_id': vaultID},
      error: error,
    );

    if (resp == null || !resp['ok']) {
      return null;
    }

    Map<String, dynamic> items = {};
    for (var i in resp['items']) {
      items[i['id']] = i;
    }

    return items;
  }

  Future<Map<String, dynamic>?> getConnectedRobots() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/workspace.robots.connected.list',
    );

    if (resp == null || !resp['ok']) {
      return null;
    }

    Map<String, dynamic> robots = {};
    for (var r in resp['robots']) {
      robots[r['id']] = r;
    }

    return robots;
  }

  Future<String> getPrivateKey(Map user, String password) async {
    final encPrivateKey = user['enc_private_key'];
    final salt = user['salt'];
    final iv = user['iv'];
    final username = user['email'];

    final algorithm = Hkdf(hmac: Hmac(Sha256()), outputLength: 32);
    final derivedSalt = await algorithm.deriveKey(
      secretKey: SecretKey(utf8.encode(username)),
      nonce: hex.decode(salt),
    );

    final pbkdf2 = Pbkdf2(
      macAlgorithm: Hmac.sha256(),
      iterations: 100000,
      bits: 256,
    );

    final mk = await pbkdf2.deriveKey(
      secretKey: SecretKey(utf8.encode(password)),
      nonce: utf8.encode(hex.encode(await derivedSalt.extractBytes())),
    );

    final _encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(await mk.extractBytes()),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final pkPem = _encrypter.decryptBytes(
      enc.Encrypted(Uint8List.fromList(hex.decode(encPrivateKey))),
      iv: enc.IV(Uint8List.fromList(hex.decode(iv))),
    );

    return utf8.decode(pkPem);
  }

  Future<Vault?> createVault(
    Map user,
    String name,
    String description,
  ) async {
    var random = Random.secure();
    var vaultKey = List<int>.generate(32, (i) => random.nextInt(256));
    var secretKey = List<int>.generate(32, (i) => random.nextInt(256))
        .toSet()
        .toList()
        .sublist(0, 16);

    secretKey = [...secretKey, ...secretKey];
    var iv = List<int>.generate(16, (i) => random.nextInt(256));
    var hexVaultKey = hex.encode(vaultKey);

    Uint8List vk = Uint8List.fromList(vaultKey);
    Uint8List sk = Uint8List.fromList(secretKey);
    var xored = vk.xor(sk);
    var xoredKey = xored.toList();

    final encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(xoredKey),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final verifier = encrypter
        .encrypt(
          hexVaultKey,
          iv: enc.IV(
            Uint8List.fromList(iv),
          ),
        )
        .base64;

    final pubPem = user['public_key'];
    final encVaultKey =
        await rsa.RSA.encryptOAEPBytes(vk, "", rsa.Hash.SHA256, pubPem);
    final encSecretKey =
        await rsa.RSA.encryptOAEPBytes(sk, "", rsa.Hash.SHA256, pubPem);

    final encVaultKeyb64 = base64.encode(encVaultKey);
    //final secretKeyb64 = base64.encode(secretKey);
    final encSecretKeyb64 = base64.encode(encSecretKey);

    final vaultIV = hex.encode(iv);

    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.create',
      body: {
        'name': name,
        'description': description,
        'enc_vault_key': encVaultKeyb64,
        'verifier': verifier,
        'iv': vaultIV,
      },
    );

    if (resp == null || !resp['ok']) {
      return null;
    }

    return Vault(
      vault: resp['vault'],
      secretKey: secretKey,
      vaultKey: vaultKey,
      encSecretKey: encSecretKeyb64,
      encVaultKey: encVaultKeyb64,
    );
  }

  Future<String> _encryptVaultItem(
    List<int> secretKey,
    List<int> vaultKey,
    List<int> iv,
    String data,
  ) async {
    var xored = Uint8List.fromList(secretKey).xor(
      Uint8List.fromList(vaultKey),
    );

    final encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(xored),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final encItem = encrypter.encrypt(
      data,
      iv: enc.IV(
        Uint8List.fromList(iv),
      ),
    );

    return hex.encode(encItem.bytes.toList());
  }

  Future<Map?> createVaultItem(
    String vaultID,
    List<int> secretKey,
    List<int> vaultKey,
    String name,
    int category,
    String data,
  ) async {
    final random = Random.secure();
    var iv = List<int>.generate(16, (i) => random.nextInt(256));

    final encVaultItem = await _encryptVaultItem(
      secretKey,
      vaultKey,
      iv,
      data,
    );

    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.items.create',
      body: {
        'vault_id': vaultID,
        'name': name,
        'category': category,
        'data': encVaultItem,
        'iv': hex.encode(iv),
      },
    );

    if (resp == null || !resp['ok']) {
      return null;
    }

    return resp['item'];
  }

  Future<bool> injectVault(
    String vaultID,
    String robotID,
    String encSecretKey,
    String encVaultKey,
  ) async {
    var resp = await apiRequest(
      method: 'POST',
      endpoint: '/v1/vaults.secret.inject',
      body: {
        'vault_id': vaultID,
        'robot_id': robotID,
        'enc_secret': encSecretKey,
        'enc_key': encVaultKey,
      },
    );

    return resp != null && resp['ok'];
  }

  Future<String?> askPassword(BuildContext context) async {
    TextEditingController _controller = TextEditingController();
    int? resp = await showDialog<int>(
      context: context,
      builder: (BuildContext context) => AlertDialog(
        title: const Text('Welcome Back!'),
        content: Column(
          mainAxisAlignment: MainAxisAlignment.start,
          mainAxisSize: MainAxisSize.min,
          children: [
            Row(
              children: [
                Text(
                  'Password',
                  style: GoogleFonts.openSans(
                    fontSize: 14,
                    fontWeight: FontWeight.w600,
                  ),
                ),
              ],
            ),
            SizedBox(
              width: 240,
              child: TextFormField(
                obscureText: true,
                enableSuggestions: false,
                autocorrect: false,
                style: GoogleFonts.openSans(
                  fontSize: 14,
                  fontWeight: FontWeight.w500,
                ),
                decoration: const InputDecoration(
                  icon: Icon(
                    Icons.password,
                  ),
                  enabledBorder: UnderlineInputBorder(
                    borderSide: BorderSide(),
                  ),
                  hintText: 'Your Password',
                ),
                controller: _controller,
              ),
            ),
          ],
        ),
        actions: [
          ElevatedButton(
            style: ButtonStyle(
              padding: MaterialStateProperty.all(
                const EdgeInsets.all(16),
              ),
              backgroundColor: MaterialStateProperty.all(Colors.blue),
              shape: MaterialStateProperty.all(
                RoundedRectangleBorder(
                  borderRadius: BorderRadius.circular(8),
                ),
              ),
            ),
            onPressed: () {
              Navigator.pop(context, 0);
            },
            child: Text(
              'Next',
              style: GoogleFonts.openSans(
                fontSize: 14,
                fontWeight: FontWeight.w600,
                color: Colors.white,
              ),
            ),
          ),
        ],
      ),
    );

    return resp == null ? null : _controller.text;
  }

  Future<String?> askVaultSecret(BuildContext context, Vault vault) async {
    TextEditingController _controller = TextEditingController();
    int? resp = await showDialog<int>(
      context: context,
      builder: (BuildContext context) => AlertDialog(
        title: const Text('Unlock Vault'),
        content: Column(
          mainAxisAlignment: MainAxisAlignment.start,
          mainAxisSize: MainAxisSize.min,
          children: [
            Row(
              children: [
                Text(
                  'Vault Secret Key',
                  style: GoogleFonts.openSans(
                    fontSize: 14,
                    fontWeight: FontWeight.w600,
                  ),
                ),
              ],
            ),
            SizedBox(
              width: 240,
              child: TextFormField(
                obscureText: true,
                enableSuggestions: false,
                autocorrect: false,
                style: GoogleFonts.openSans(
                  fontSize: 14,
                  fontWeight: FontWeight.w500,
                ),
                decoration: const InputDecoration(
                  icon: Icon(
                    Icons.password,
                  ),
                  enabledBorder: UnderlineInputBorder(
                    borderSide: BorderSide(),
                  ),
                  hintText: 'Type the secret key here',
                ),
                controller: _controller,
              ),
            ),
          ],
        ),
        actions: [
          ElevatedButton(
            style: ButtonStyle(
              padding: MaterialStateProperty.all(
                const EdgeInsets.all(16),
              ),
              backgroundColor: MaterialStateProperty.all(Colors.blue),
              shape: MaterialStateProperty.all(
                RoundedRectangleBorder(
                  borderRadius: BorderRadius.circular(8),
                ),
              ),
            ),
            onPressed: () {
              Navigator.pop(context, 0);
            },
            child: Text(
              'Unlock',
              style: GoogleFonts.openSans(
                fontSize: 14,
                fontWeight: FontWeight.w600,
                color: Colors.white,
              ),
            ),
          ),
        ],
      ),
    );

    return resp == null ? null : _controller.text;
  }
}

class Config {
  String _id = "";
  //final String namespace;
  //final String token;
  String _userID = "";
  //final String repositoryID;
  final String? workspace;
  String? robotID;
  String? flowID;
  late Map _profile;

  Config(
      { //@required this.id,
      //@required this.namespace,
      //@required this.token,
      //@required this.userID,
      //@required this.repositoryID,
      this.workspace}) {
    if (kDebugMode) {
      this._id = _getCookie("app_id") ?? "7c07607d-45e7-4fc2-b1dc-7806edd1c39e";
    } else {
      List<String> parts = Uri.base.toString().split("/");
      for (String p in parts.reversed) {
        try {
          Uuid.parse(p);
          this._id = p;
          break;
        } catch (e) {}
      }
    }

    var cookie = _getCookie("profile");
    if (cookie != null) {
      this._profile =
          jsonDecode(Uri.decodeFull(cookie)) ?? "{}" as Map<dynamic, dynamic>;
      this._userID = this._profile["id"] ?? "";
    }
  }
}

String? _getCookie(String cname) {
  Map _cookies = {};
  html.document.cookie
      ?.split(";")
      .map(
        (c) => c.split("="),
      )
      .where((c) => c.length > 1)
      .forEach(
        (c) => _cookies[c[0].trim()] = c[1],
      );

  return _cookies[cname];
}
